;+# ***** BEGIN COPYRIGHT BLOCK *****
;+# Copyright (C) 2005 Red Hat, Inc.
;+# All rights reserved.
;+#
;+# This library is free software; you can redistribute it and/or
;+# modify it under the terms of the GNU Lesser General Public
;+# License as published by the Free Software Foundation version
;+# 2.1 of the License.
;+#
;+# This library is distributed in the hope that it will be useful,
;+# but WITHOUT ANY WARRANTY; without even the implied warranty of
;+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;+# Lesser General Public License for more details.
;+#
;+# You should have received a copy of the GNU Lesser General Public
;+# License along with this library; if not, write to the Free Software
;+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
;+# ***** END COPYRIGHT BLOCK *****

;+#
;+# Coreconf smart .def file.
;+#  This .def file has encoded in it enough information so it can easily be
;+#  converted to an appropriate export symbol file an just about any platform.
;+#
;+#  To add new symbols to this file, be sure to enclose those symbols in the
;+#  appropriate shared library version number. These version numbers are parsed
;+#  by Linux and Solaris. These commands should be prefixed with the 
;+#  appropriate ';+'
;+#
;+#  general comments are prefixed by ';+#'
;+#  
;+#  Windows & OS/2 key words should not be prefixed. The would contain the
;+#  the line ';-' somewhere  in the line. This file will be fed 'AS IS' to
;+#  Windows and OS/2 with no processing. Anything following a ';' on these 
;+#  platforms are already treated as a comment.
;+#
;+#  As meantioned above, Linux and Solaris commands should be prefixed with ';+'
;+#  The ;+ will be removed, as well as any line that has ";-". The " DATA " 
;+#  keyword will also be removed. The '#' is a comment on these platforms.
;+#
;+#  A simple list of export symbols (needed for some platforms like AIX or 
;+#  libtool can be generated by running the following command:
;+#      grep -v ';+' $(DEF_FILE) | grep -v ';-' | \
;+#             sed -e 's; DATA ;;' -e 's,;;,,' -e 's,;.*,,'  > $(SYM_LIST)
;+#
;+#  Some platforms take symbols in the form of a sym list with some prefix,
;+#  for instance the MAC symbol table can be processed with the above command
;+#  plus an extra -e 's,^,_,' to the sed command above. HP symbols can be
;+#  processed by adding an extra -e 's,^,-e ,'.
;+#
;+COOLKEY_2.10 {    # Cool Key 2.10 is the PKCS #11 spec version
;+    global:
LIBRARY coolkey ;-
EXPORTS ;-
C_GetFunctionList
;+# Make this function like a real PKCS #11 module as well
C_CancelFunction;
C_CloseAllSessions;
C_CloseSession;
C_CopyObject;
C_CreateObject;
C_Decrypt;
C_DecryptDigestUpdate;
C_DecryptFinal;
C_DecryptInit;
C_DecryptUpdate;
C_DecryptVerifyUpdate;
C_DeriveKey;
C_DestroyObject;
C_Digest;
C_DigestEncryptUpdate;
C_DigestFinal;
C_DigestInit;
C_DigestKey;
C_DigestUpdate;
C_Encrypt;
C_EncryptFinal;
C_EncryptInit;
C_EncryptUpdate;
C_Finalize;
C_FindObjects;
C_FindObjectsFinal;
C_FindObjectsInit;
C_GenerateKey;
C_GenerateKeyPair;
C_GenerateRandom;
C_GetAttributeValue;
C_GetFunctionStatus;
C_GetInfo;
C_GetMechanismInfo;
C_GetMechanismList;
C_GetObjectSize;
C_GetOperationState;
C_GetSessionInfo;
C_GetSlotInfo;
C_GetSlotList;
C_GetTokenInfo;
C_Initialize;
C_InitPIN;
C_InitToken;
;+    local:
;+       *;
;+};
